import math
import numpy as np


class ParamGrid:
    __keys = []

    param_grid = None

    key_to_index_map = {}
    index_to_key_map = {}

    def __init__(self, param_grid):
        self.param_grid = param_grid

        if param_grid is None:
            raise ValueError('param_grid cannot be None!')

        param_grid_key_indices = tuple(param_grid.keys())

        for index, key in enumerate(param_grid_key_indices):
            self.__keys.append(key)
            self.key_to_index_map[key] = index
            self.index_to_key_map[index] = key

    def __len__(self):
        return len(self.param_grid)

    def keys(self):
        return tuple(self.__keys)

    def values(self):
        return self.param_grid.values()

    def get_number_of_candidates(self):
        return np.multiply.reduce([len(value) for value in self.param_grid.values()])

    def get_params_from_solution_vec(self, solution_vec):
        params = {}

        for param_key in self.param_grid:
            params[param_key] = self.__get_param_value(param_key, solution_vec)

        return params

    def __get_param_value(self, param_key, solution_vec):
        index = self.key_to_index_map[param_key]
        solution_value = solution_vec[index]
        param_value_count = len(self.param_grid[param_key])
        param_value_index = min(math.floor(solution_value / (1 / param_value_count)), param_value_count - 1)
        return self.param_grid[param_key][param_value_index]



import numpy as np

from ._param_grid import ParamGrid
from ._parameter_search_benchmark import ParameterSearchBenchmark
from ._stagnation_stopping_task import StagnationStoppingTask

from NiaPy.algorithms.basic import FireflyAlgorithm, BatAlgorithm, GreyWolfOptimizer

from sklearn.model_selection._search import BaseSearchCV

SUPPORTED_ALGORITHMS = ['fa', 'gwo', 'ba']


class NatureInspiredSearchCV(BaseSearchCV):
    __param_grid = None
    __algorithm = None

    def _run_search(self, evaluate_candidates):
        self.__print_run_search_log()

        benchmark = ParameterSearchBenchmark(evaluate_candidates, self.__param_grid)

        task = StagnationStoppingTask(
            length=len(self.__param_grid),
            n_gen=self.__n_gen,
            max_stagnating_gen=self.__max_stagnating_gen,
            benchmark=benchmark
        )

        self.__algorithm.run(task=task)
        # invert scores in the optimization logs
        self.optimization_logs_ = [(log[0], log[1] * -1) for log in task.optimization_logs_]

    def __init__(self, estimator, param_grid, algorithm='ba', population_size=15, max_n_gen=25,
                 max_stagnating_gen=2, scoring=None, n_jobs=None, iid='deprecated', refit=True, cv=None, verbose=0,
                 pre_dispatch='2*n_jobs', error_score=np.nan, return_train_score=True):
        super().__init__(estimator, scoring, n_jobs, iid, refit, cv, verbose, pre_dispatch, error_score,
                         return_train_score)

        self.__n_gen = max_n_gen
        self.__max_stagnating_gen = max_stagnating_gen
        self.__param_grid = ParamGrid(param_grid)
        self.__algorithm = self.__get_algorithm(algorithm, population_size)

        self.optimization_logs_ = None

    def __print_run_search_log(self):
        candidates = self.__param_grid.get_number_of_candidates()

        if self.cv is not None:
            print(f'Fitting {self.cv} folds for some of the {candidates} candidates, '
                  f'which might total in {candidates * self.cv} fits')
        else:
            print(f'Fitting at most {candidates} candidates')

    @staticmethod
    def __get_algorithm(algorithm, population_size):
        if isinstance(algorithm, str):
            if algorithm not in SUPPORTED_ALGORITHMS:
                raise ValueError(f'"{algorithm}" is not in supported algorithms: {", ".join(SUPPORTED_ALGORITHMS)}')

            algorithm_obj = None

            if algorithm == 'fa':
                algorithm_obj = FireflyAlgorithm()
            elif algorithm == 'ba':
                algorithm_obj = BatAlgorithm()
            elif algorithm == 'gwo':
                algorithm_obj = GreyWolfOptimizer()

            algorithm_obj.setParameters(NP=population_size)

            return algorithm_obj

        return algorithm



from NiaPy.benchmarks import Benchmark


class ParameterSearchBenchmark(Benchmark):
    evaluation_count = 0

    cache = {}

    def __init__(self, evaluate_candidates, param_grid):
        self.evaluate_candidates = evaluate_candidates
        self.param_grid = param_grid

        Benchmark.__init__(self, 0, 1)

    def get_cache_key(self, params):
        keys = self.param_grid.keys()
        return tuple([params[key] for key in keys])

    def get_cached_score(self, params):
        cache_key = self.get_cache_key(params)

        if cache_key in self.cache:
            return self.cache[cache_key]

        return None

    def set_cached_score(self, params, score):
        cache_key = self.get_cache_key(params)
        self.cache[cache_key] = score

    def function(self):
        def evaluate(_, solution_vec):
            params = self.param_grid.get_params_from_solution_vec(solution_vec)

            score = self.get_cached_score(params)

            if score is None:
                cv_results = self.evaluate_candidates([params])
                mean_test_score = cv_results['mean_test_score']
                # we need to invert the score since we are doing
                # a minimization task
                score = -mean_test_score[self.evaluation_count]
                self.evaluation_count += 1
                self.set_cached_score(params, score)

            return score

        return evaluate



import seaborn as sns
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt


def score_by_generation_lineplot(nia_search, **kwargs):
    metrics = kwargs.get('metrics', ['min', 'max', 'mean', 'median'])

    df = pd.DataFrame(nia_search.optimization_logs_, columns=['generation', 'score'])
    score_df = df.groupby('generation')\
        .agg({'score': [np.min, np.max, np.mean, np.median]})\
        .reset_index()\
        .score\
        .rename(columns={'amax': 'max', 'amin': 'min'})

    data = []

    for metric in score_df.columns:
        for generation, score in enumerate(score_df[metric]):
            data.append((generation, metric, score))

    new_df = pd.DataFrame(data, columns=['generation', 'metric', 'score'])
    new_df = new_df[new_df['metric'].isin(metrics)]

    ax = kwargs.get('ax', None)
    ylim = kwargs.get('ylim', None)

    if ax is None:
        _, ax = plt.subplots()

    if ylim is not None:
        ax.set(ylim=ylim)

    return sns.lineplot(x="generation", y="score", data=new_df, hue='metric')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sklearn_nature_inspired_algorithms',
 'sklearn_nature_inspired_algorithms.helpers',
 'sklearn_nature_inspired_algorithms.model_selection']

package_data = \
{'': ['*']}

install_requires = \
['NiaPy==2.0.0rc10',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.4,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'scikit-learn>=0.22.2,<0.23.0',
 'seaborn>=0.10.1,<0.11.0',
 'toml>=0.9,<0.10']

setup_kwargs = {
    'name': 'sklearn-nature-inspired-algorithms',
    'version': '0.2.2',
    'description': 'Search using nature inspired algorithms over specified parameter values for an sklearn estimator.',
    'long_description': "# Nature Inspired Algorithms for scikit-learn\n\n[![CI](https://github.com/timzatko/Sklearn-Nature-Inspired-Algorithms/workflows/CI/badge.svg?branch=master)](https://github.com/timzatko/Sklearn-Nature-Inspired-Algorithms/actions?query=workflow:CI+branch:master)\n[![Maintainability](https://api.codeclimate.com/v1/badges/ed99e5c765bf5c95d716/maintainability)](https://codeclimate.com/github/timzatko/Sklearn-Nature-Inspired-Algorithms/maintainability)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/sklearn-nature-inspired-algorithms)\n[![PyPI version](https://badge.fury.io/py/sklearn-nature-inspired-algorithms.svg)](https://pypi.org/project/sklearn-nature-inspired-algorithms/)\n[![PyPI downloads](https://img.shields.io/pypi/dm/sklearn-nature-inspired-algorithms)](https://pypi.org/project/sklearn-nature-inspired-algorithms/)\n \nNature inspired algorithms for hyper-parameter tuning of [scikit-learn](https://github.com/scikit-learn/scikit-learn) models. This package uses algorithms implementation from [NiaPy](https://github.com/NiaOrg/NiaPy). \n\n## Installation\n\n```shell script\npip install sklearn-nature-inspired-algorithms\n```\n\n## Usage\n\nThe usage is similar to using sklearn's `GridSearchCV`.\n\n```python\nfrom sklearn_nature_inspired_algorithms.model_selection import NatureInspiredSearchCV\nfrom sklearn.ensemble import RandomForestClassifier\n\nparam_grid = { \n    'n_estimators': range(20, 100, 20), \n    'max_depth': range(2, 20, 2),\n    'min_samples_split': range(2, 20, 2), \n}\n\nclf = RandomForestClassifier(random_state=42)\n\nnia_search = NatureInspiredSearchCV(\n    clf,\n    param_grid,\n    algorithm='ba', # bat algorithm\n    population_size='15',\n    max_n_gen=30,\n    max_stagnating_gen=2,\n)\n\nnia_search.fit(X_train, y_train)\n```\n\nJupyter notebooks with full examples are available in [here](examples/notebooks).\n\n### Using custom nature inspired algorithm\n\nIf you do not want to use ony of the pre-defined algorithm configurations, you can use any algorithm from the  [NiaPy](https://github.com/NiaOrg/NiaPy) collection.\nThis will allow you to have more control of the algorithm behaviour. \nRefer to their [documentation](https://niapy.readthedocs.io/en/latest/) and [examples](https://github.com/NiaOrg/NiaPy/tree/master/examples) for the usage. \n\n```python\nfrom NiaPy.algorithms.basic import GeneticAlgorithm\n\nalgorithm = GeneticAlgorithm()\nalgorithm.setParameters(NP=50, Ts=5, Mr=0.25)\n\nnia_search = NatureInspiredSearchCV(\n    clf,\n    param_grid,\n    algorithm=algorithm,\n    population_size='15',\n    max_n_gen=30,\n    max_stagnating_gen=2,\n)\n\nnia_search.fit(X_train, y_train)\n```\n\n## Contributing \n\nDetailed information on the contribution guidelines are in the [CONTRIBUTING.md](./CONTRIBUTING.md).",
    'author': 'Timotej Zatko',
    'author_email': 'timi.zatko@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/timzatko/Sklearn-Nature-Inspired-Algorithms',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
